(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2013-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uhistory.ml,v 1.15 2014-03-07 23:53:28 fclement Exp $ *)

module Nat = Unatural;;
module Cmd_opt = Ucommand_option;;

let string_of_list s_of l =
  let len = Nat.to_int (Ulist.length l) in
  Ulist.fold_left_i
    (fun i accu x ->
      match i with
      | 0 -> accu ^ s_of x
      | n when n = len - 1 -> accu ^ ", and " ^ s_of x
      | _ -> accu ^ ", " ^ s_of x)
    "" l
;;

let h_qual = Upervasives.qualify "Uhistory";;

let h_failwith = Upervasives.failwith_qual h_qual;;

open Uhistory_types;;
open Uhistory_types_print;;
open Uhistory_spec;;

module Make =
  functor (Config : CONFIGURATION) ->
  functor (St : STATE) ->
struct
  type config = Config.t;;
  type state = St.t;;
  
  type t = (config, state) Uhistory_types.t;;
  
  (* Write directory name. *)
  
  let default_write_directory_name = ".";;
  
  let { Uaref.get = get_write_directory_name;
        set = set_write_directory_name; } =
    Uaref.make_aref_single "write directory name" default_write_directory_name
  ;;
  
  let option_write_directory_name =
    {
     Cmd_opt.key = "-write-directory-name";
     spec = Arg.String set_write_directory_name;
     doc =
     Printf.sprintf
       "<dname>\n\
        \tSet write directory name to <dname>.\n\
        \tDefault is %S."
      (get_write_directory_name ());
    }
  ;;
  
  let marshal_save_restore () =
    (fun fname (x : 'a) ->
      let oc = open_out_bin fname in
      Marshal.to_channel oc x [ Marshal.Closures; ];
      close_out oc;),
    (fun fname ->
      let ic = open_in_bin fname in
      let x = Marshal.from_channel ic in
      close_in ic;
      (x : 'a))
  ;;
  
  let (save_history, restore_history) = marshal_save_restore ();;
  
  let head_state { h_list = ({ state = sti; _; }, _); _; } = sti;;
  let head_state_number { h_list = ({ state_number = i; _; }, _); _; } = i;;
  
  let make_trivial_history_list i sti =
    ({ state_number = i; state = sti; }, [])
  ;;
  
  let last_history_step i sti h =
    { h with h_list = make_trivial_history_list i sti; }
  ;;
  
  let full_history_step i sti ({ h_list = hlim1; _; } as h) =
    { h with h_list = Ulist1.cons { state_number = i; state = sti; } hlim1; }
  ;;
  
  let filename_of_basename bname h =
    let i = head_state_number h in
    let fname = Printf.sprintf "%s_%d.obj" bname (Nat.to_int i) in
    let dname = get_write_directory_name () in
    Filename.concat dname fname
  ;;
  
  let checkpoint ckp h =
    match ckp with
    | None -> ""
    | Some bname ->
      let fname = filename_of_basename bname h in
      save_history fname h;
      fname
  ;;
  
  let make_history_step i sti ({ h_mode = (hs_mode, ckp); _; } as him1) =
    let hi =
      match hs_mode with
      | Last -> last_history_step i sti him1
      | Full -> full_history_step i sti him1 in
    ignore (checkpoint ckp hi);
    hi
  ;;
  
  let same_head_state h1 h2 =
    St.same (head_state h1) (head_state h2)
  ;;
  
  (* Printer shortcuts. *)
  
  let print ppf h = print_t Config.print St.print ppf h;;
  
  let print_head_h_state ppf h =
    print_h_state St.print ppf (Ulist1.hd h.h_list)
  ;;
  
  (* Cold/warm start. *)
  
  (* Start mode. *)
  
  let { Uaref.get = get_start_mode;
        set = set_start_mode; } =
    Uaref.make_aref_single "start mode" Cold_start
  ;;
  
  let start_mode_of_string = function
    | "cold" | "c" -> Cold_start
    | "warm" | "w" -> Warm_start
    | s ->
        h_failwith "start_mode_of_string"
          (Printf.sprintf "Unknown start_mode %S" s)
  ;;
  let set_start_mode_from_string s = set_start_mode (start_mode_of_string s);;
  
  let start_mode_to_string = function
    | Cold_start -> "cold"
    | Warm_start -> "warm"
  ;;
  let start_mode_to_quoted_string x =
    Printf.sprintf "%S" (start_mode_to_string x)
  ;;
  let print_default_start_mode () =
    start_mode_to_quoted_string (get_start_mode ())
  ;;
  let print_all_start_modes () =
    string_of_list start_mode_to_quoted_string [ Cold_start; Warm_start; ]
  ;;
  
  let option_start_mode =
    {
     Cmd_opt.key = "-start-mode";
     spec = Arg.String set_start_mode_from_string;
     doc =
       Printf.sprintf
         "<str>\n\
          \tSet start mode to <str>.\n\
          \tValid start modes are: %t.\n\
          \tWarm start mode needs option \
            \"-restore-history-file-name\" to be set.\n\
          \tDefault start mode is %t."
         print_all_start_modes
         print_default_start_mode;
    }
  ;;
  
  let { Uaref.get = get_restore_history_file_name;
        set = set_restore_history_file_name; } =
    Uaref.make_aref_unique "restore history file name" ""
  ;;
  
  let option_restore_history_file_name =
    {
     Cmd_opt.key = "-restore-history-file-name";
     spec = Arg.String set_restore_history_file_name;
     doc =
       Printf.sprintf
         "<fname>\n\
          \tSet the file name from which restoring history to <fname>.\n\
          \tFor warm restart only.";
    }
  ;;
  
  let { Uaref.get = get_save_history_file_name;
        set = set_save_history_file_name; } =
    Uaref.make_aref_single "save history file name" ""
  ;;
  
  let option_save_history_file_name =
    {
     Cmd_opt.key = "-save-history-file-name";
     spec = Arg.String set_save_history_file_name;
     doc =
       Printf.sprintf
         "<fname>\n\
          \tSet the file name to which saving history to <fname>.\n\
          \tFor further warm restart. An empty value means no saving.";
    }
  ;;
  
  (* History mode. *)
  
  let { Uaref.get = get_history_step_mode;
        set = set_history_step_mode; } =
    Uaref.make_aref_single "history step mode" Last
  ;;
  
  let history_step_mode_of_string = function
    | "last" | "l" -> Last
    | "full" | "f" -> Full
    | s ->
        h_failwith "history_step_mode_of_string"
          (Printf.sprintf "Unknown history step mode %S" s)
  ;;
  let set_history_step_mode_from_string s =
    set_history_step_mode (history_step_mode_of_string s)
  ;;
  
  let history_step_mode_to_string = function
    | Last -> "last"
    | Full -> "full"
  ;;
  let history_step_mode_to_quoted_string x =
    Printf.sprintf "%S" (history_step_mode_to_string x)
  ;;
  let print_default_history_step_mode () =
    history_step_mode_to_quoted_string (get_history_step_mode ())
  ;;
  let print_all_history_step_modes () =
    string_of_list history_step_mode_to_quoted_string [ Last; Full; ]
  ;;
  
  let option_history_step_mode =
    {
     Cmd_opt.key = "-history-step-mode";
     spec = Arg.String set_history_step_mode_from_string;
     doc =
       Printf.sprintf
         "<str>\n\
          \tSet history step mode to <str>.\n\
          \tValid history step modes are: %t.\n\
          \tDefault history step mode is %t."
         print_all_history_step_modes
         print_default_history_step_mode;
    }
  ;;
  
  let { Uaref.get = get_checkpointing;
        set = set_checkpointing; } =
    Uaref.make_aref_single "checkpointing" ""
  ;;
  
  let option_checkpointing =
    {
     Cmd_opt.key = "-checkpointing";
     spec = Arg.String set_checkpointing;
     doc =
       Printf.sprintf
         "<str>\n\
          \tSet basename of checkpointing files to <str>.\n\
          \tAn empty value means no checkpointing.\n\
          \tDefault is no checkpointing.";
    }
  ;;
  
  let get_history_mode () =
    let hsm = get_history_step_mode () in
    let ckp =
      match get_checkpointing () with
      | "" -> None
      | s -> Some s in
    (hsm, ckp)
  ;;
  
  (* Command-line options. *)
  
  let regular_opt rco =
    Cmd_opt.with_auto_short_key
      (Cmd_opt.with_auto_gnu_key
        (Cmd_opt.of_regular rco))
  ;;
  
  let user_add_all l =
    Cmd_opt.user_add_all (Ulist.map regular_opt l)
  ;;
  
  let phase_start_mode = Cmd_opt.Phase "start_mode";;
  
  let add_history_options () =
    user_add_all
      [
       option_start_mode;
       option_restore_history_file_name;
       option_write_directory_name;
       option_save_history_file_name;
       option_history_step_mode;
       option_checkpointing;
      ];
    Cmd_opt.phase_add_all
      phase_start_mode
      [ "-start-mode";
        "-restore-history-file-name";
        "-write-directory-name"; ];
  ;;
  add_history_options ();;
  
  let initialization () =
    Cmd_opt.parse_phase phase_start_mode;
    let rec loop = function
      | Cold_start ->
        Cmd_opt.parse ();
        Config.cold_set ();
        let h_state_0 = St.init () in
        let h0 =
          {
           configuration = Config.get ();
           h_mode = get_history_mode ();
           h_list = make_trivial_history_list Nat._0_ h_state_0;
          } in
        let bname =
          match get_checkpointing () with
          | "" -> "cold"
          | s -> s in
        let fname = checkpoint (Some bname) h0 in
        set_restore_history_file_name fname;
        loop Warm_start;
      | Warm_start ->
        let h0 = restore_history (get_restore_history_file_name ()) in
        Cmd_opt.parse ();
        Config.warm_set h0.configuration;
        let x = { h0 with configuration = Config.get (); } in
        x in
    loop (get_start_mode ())
  ;;
  
  let finalization hn =
    let h_state_n = head_state hn in
    let dname = get_write_directory_name () in
    St.final h_state_n;
    match get_save_history_file_name () with
    | "" -> ()
    | fname -> save_history (Filename.concat dname fname) hn
  ;;
end
;;
